unit CubeMath;

interface


uses

   System.UITypes, System.UIConsts, System.SysUtils, System.Classes, System.Types, System.Math,

{$ifdef USE_FMX}
  FMX.Types, FMX.Graphics;
{$else}
  Vcl.Graphics;
{$endif}


  // Numbering scheme for the cube of faces of panels of cells ... print this, cut it out and fold the cube :)
  //
  //                 +-------+-------+
  //                 | 0   1 | 0   1 |
  //                 |   0   |   1   |
  //                 | 2   3 | 2   3 |
  //                 +------ 0 ------+
  //                 | 0   1 | 0   1 |
  //                 |   2   |   3   |
  //                 | 2   3 | 2   3 |               |
  // +-------+-------+-------+-------+-------+-------+-------+-------+
  // | 0   1 | 0   1 | 0   1 | 0   1 | 0   1 | 0   1 | 0   1 | 0   1 |
  // |   0   |   1   |   0   |   1   |   0   |   1   |   0   |   1   |
  // | 2   3 | 2   3 | 2   3 | 2   3 | 2   3 | 2   3 | 2   3 | 2   3 |
  // +------ 1 ------+------ 2 ------+------ 3 ------+------ 4 ------+
  // | 0   1 | 0   1 | 0   1 | 0   1 | 0   1 | 0   1 | 0   1 | 0   1 |
  // |   2   |   3   |   2   |   3   |   2   |   3   |   2   |   3   |
  // | 2   3 | 2   3 | 2   3 | 2   3 | 2   3 | 2   3 | 2   3 | 2   3 |
  // +-------+-------+-------+-------+-------+-------+-------+-------+
  //                 | 0   1 | 0   1 |               |
  //                 |   0   |   1   |
  //                 | 2   3 | 2   3 |
  //                 +------ 5 ------+
  //                 | 0   1 | 0   1 |
  //                 |   2   |   3   |
  //                 | 2   3 | 2   3 |
  //                 +-------+-------+
  //
  // a cell is a single LED
  // a panel is a set of four cells
  // a face is a set of four panels
  // the cube is a set of six faces


const

  RULE_COUNT     = 8;
  CELLS_IN_PANEL = 4;
  PANELS_IN_FACE = 4;
  FACES_IN_CUBE  = 6;


type

  TOnGetbackground = function( const aSender: TObject): string of object;

  TAction = (
    acNothing,
    acDie    ,
    acLive
  );

  TBasicRule = array[ 0 .. RULE_COUNT - 1] of TAction;
  TRule      = array[ Boolean] of TBasicRule;

  TCubeCell = class
  private
    FFace           : Integer;
    FPanel          : Integer;
    FCell           : Integer;
    FAlive          : Boolean;
    FNewAlive       : Boolean;
    FRed            : Boolean;
    FNewRed         : Boolean;
    FRuleGreen      : TRule;
    FRuleRed        : TRule;
    FUse1ForMissing : Boolean;
    FShowNumbers    : Boolean;
    FNeighbours     : array of TCubeCell;
  private
    function    GetRuleGreen: TRule;
    procedure   SetRuleGreen( const aValue: TRule);
    function    GetRuleRed  : TRule;
    procedure   SetRuleRed  ( const aValue: TRule);
    function    GetState    : Integer;
    procedure   SetState    ( aValue: Integer);
    function    GetNeighbourCount : Integer;
    function    GetNeighbour( anIndex: Integer): TCubeCell;
    function    DumpNeighbours: string;
  protected
    procedure   PaintOn( const aCanvas: TCanvas; const aBounds: TRect; EraseBackGround: Boolean);
    function    MouseToCell( anX, anY: Integer; const aBounds: TRect): TCubeCell;
  public
    constructor Create( aFace, aPanel, aCell: Integer);
    procedure   ConnectTo( const aCell: TCubeCell);
    function    CountLivingNeighbours: Integer;
    procedure   Process( UseRed: Boolean);
    procedure   Proceed;
    function    Dump: string;
    procedure   Clear;
  public
    property    RuleGreen                    : TRule     read GetRuleGreen      write SetRuleGreen;
    property    RuleRed                      : TRule     read GetRuleRed        write SetRuleRed;
    property    Alive                        : Boolean   read FAlive            write FAlive;
    property    Red                          : Boolean   read FRed              write FRed;
    property    State                        : Integer   read GetState          write SetState;
    property    Use1ForMissing               : Boolean   read FUse1ForMissing   write FUse1ForMissing;
    property    ShowNumbers                  : Boolean   read FShowNumbers      write FShowNumbers;
    property    Face                         : Integer   read FFace;
    property    Panel                        : Integer   read FPanel;
    property    Cell                         : Integer   read FCell;
    property    NeighbourCount               : Integer   read GetNeighbourCount;
    property    Neighbour[ anIndex: Integer] : TCubeCell read GetNeighbour;
  end;

  TCubeFace  = class;

  TCubePanel = class
  private
    FPArent      : TCubeFace;
    FCells       : array[ 0 .. CELLS_IN_PANEL - 1] of TCubeCell;
    FFace        : Integer;
    FPanel       : Integer;
    FShowNumbers : Boolean;
  {$ifdef USE_FMX}
    FBackGround  : TBitmap;
    FBackGroundR : TBitmap; // A 90 degree rotated bitmap used on faces 2 and 4
  {$endif}
  private
    function    GetCell( aCell: Integer): TCubeCell;
    procedure   SetCell( aCell: Integer; const aValue: TCubeCell);
    function    GetRuleGreen     : TRule;
    procedure   SetRuleGreen     ( const aValue: TRule);
    function    GetRuleRed       : TRule;
    procedure   SetRuleRed       ( const aValue: TRule);
    function    GetState         : string;
    procedure   SetState         ( const aValue: string);
    procedure   SetUse1ForMissing( aValue: Boolean);
    procedure   SetShowNumbers   ( aValue: Boolean);
    procedure   FixLocalConnections;
  protected
    procedure   PaintOn( const aCanvas: TCanvas; const aBounds: TRect);
    function    MouseToCell( anX, anY: Integer; const aBounds: TRect): TCubeCell;
    function    Getbackground: string;
  public
    constructor Create( const aParent: TCubeFace; aFace, aPanel: Integer);
    destructor  Destroy;                                                                                       override;
    procedure   Process( UseRed: Boolean);
    procedure   Proceed;
    function    Dump: string;
    procedure   Clear;
  public
    property    RuleGreen             : TRule     read GetRuleGreen write SetRuleGreen;
    property    RuleRed               : TRule     read GetRuleRed   write SetRuleRed;
    property    Cell[ aCell: Integer] : TCubeCell read GetCell      write SetCell;                              default;
    property    State                 : string    read GetState     write SetState;
    property    Use1ForMissing        : Boolean                     write SetUse1ForMissing;
    property    ShowNumbers           : Boolean   read FShowNumbers write SetShowNumbers;
    property    Face                  : Integer   read FFace;
    property    Panel                 : Integer   read FPanel;
  end;


  TCubeLife = class;

  TCubeFace = class
  private
    FParent      : TCubeLife;
    FPanels      : array[ 0 .. PANELS_IN_FACE - 1] of TCubePanel;
    FFace        : Integer;
    FShowNumbers : Boolean;
  private
    function    GetPanel( aPanel: Integer): TCubePanel;
    procedure   SetPanel( aPanel: Integer; const aValue: TCubePanel);
    function    GetCell ( aPanel, aCell: Integer): TCubeCell;
    procedure   SetCell ( aPanel, aCell: Integer; const aValue: TCubeCell);
    function    GetRuleGreen     : TRule;
    procedure   SetRuleGreen     ( const aValue: TRule);
    function    GetRuleRed       : TRule;
    procedure   SetRuleRed       ( const aValue: TRule);
    function    GetState         : string;
    procedure   SetState         ( const aValue: string);
    procedure   SetUse1ForMissing( aValue: Boolean);
    procedure   SetShowNumbers   ( aValue: Boolean);
    procedure   FixLocalConnections;
  protected
    procedure   PaintOn( const aCanvas: TCanvas; const aBounds: TRect);
    function    MouseToCell( anX, anY: Integer; const aBounds: TRect): TCubeCell;
    function    Getbackground: string;
  public
    constructor Create( const aParent: TCubeLife; aFace: Integer);
    destructor  Destroy;                                                                                       override;
    procedure   Process( UseRed: Boolean);
    procedure   Proceed;
    function    Dump: string;
    procedure   Clear;
  public
    property    RuleGreen                      : TRule      read GetRuleGreen  write SetRuleGreen;
    property    RuleRed                        : TRule      read GetRuleRed    write SetRuleRed;
    property    Panel[ aPanel: Integer       ] : TCubePanel read GetPanel      write SetPanel;                  default;
    property    Cell [ aPanel, aCell: Integer] : TCubeCell  read GetCell       write SetCell;
    property    State                          : string     read GetState      write SetState;
    property    Use1ForMissing                 : Boolean                       write SetUse1ForMissing;
    property    ShowNumbers                    : Boolean    read FShowNumbers  write SetShowNumbers;
    property    Face                           : Integer    read FFace;
  end;

  TCubeHistory = class
  private
    FData       : array of string;
    FFillCount  : Integer;
    FFreeCount  : Integer;
    FPointer    : Integer;
  public
    constructor Create( aSize: Integer);
    destructor  Destroy;                                                                                       override;
    procedure   Clear;
    procedure   Push( const aValue: string);
    function    Pop: string;
  public
    property    FillCount: Integer read FFillCount;
  end;


  TCubeLife = class
  private
    FFaces           : array[ 0 .. FACES_IN_CUBE - 1] of TCubeFace;
    FRuleStrGreen    : string;
    FRuleStrRed      : string;
    FHistory         : TCubeHistory;
    FUseRedRule      : Boolean;
    FUse1ForMissing  : Boolean;
    FShowNumbers     : Boolean;
    FOnGetBackground : TOnGetBackground;
  private
    function    GetFace          ( aFace: Integer): TCubeFace;
    procedure   SetFace          ( aFace: Integer; const aValue: TCubeFace);
    function    GetPanel         ( aFace, aPanel: Integer): TCubePanel;
    procedure   SetPanel         ( aFace, aPanel: Integer; const aValue: TCubePanel);
    function    GetCell          ( aFace, aPanel, aCell: Integer): TCubeCell;
    procedure   SetCell          ( aFace, aPanel, aCell: Integer; const aValue: TCubeCell);
    function    GetFaceCount     : Integer;
    function    GetPanelCount    ( aFace        : Integer): Integer;
    function    GetCellCount     ( aFace, aPanel: Integer): Integer;
    function    GetRuleGreen     : TRule;
    procedure   SetRuleGreen     ( const aValue: TRule);
    function    GetRuleRed       : TRule;
    procedure   SetRuleRed       ( const aValue: TRule);
    procedure   SetRuleStrGreen  ( const aValue: string);
    procedure   SetRuleStrRed    ( const aValue: string);
    function    GetPattern       : string;
    procedure   SetPattern       ( const aValue: string);
    function    GetState         : string;
    procedure   SetState         ( const aValue: string);
    function    GetHistoryCount  : Integer;
    procedure   SetUse1ForMissing( aValue: Boolean);
    procedure   SetShowNumbers   ( aValue: Boolean);
    procedure   FixLocalConnections;
  protected
    function    Getbackground: string;
  public
    constructor Create( aHistorySize: Integer);
    destructor  Destroy;                                                                                       override;
    procedure   Process;
    procedure   Unprocess;
    function    Dump: string;
    procedure   PaintFaceOn( aFace: Integer; const aCanvas: TCanvas; const aBounds: TRect);
    procedure   PaintOn( const aCanvas: TCanvas; const aBounds: TRect);
    procedure   Clear;
    function    MouseToCell( anX, anY: Integer; const aBounds: TRect): TCubeCell;
    function    MouseToFaceCell( aFace: Integer; anX, anY: Integer; const aBounds: TRect): TCubeCell;
    procedure   LoadPattern( const aFileName: string);
    procedure   SavePattern( const aFileName: string);
    procedure   ResetHistory;
  public
    property    RuleGreen                             : TRule            read GetRuleGreen     write SetRuleGreen;
    property    RuleRed                               : TRule            read GetRuleRed       write SetRuleRed;
    property    RuleStrGreen                          : String           read FRuleStrGreen    write SetRuleStrGreen;
    property    RuleStrRed                            : String           read FRuleStrRed      write SetRuleStrRed;
    property    FaceCount                             : Integer          read GetFaceCount;
    property    PanelCount[ aFace: Integer]           : Integer          read GetPanelCount;
    property    CellCount[ aFace, aPanel: Integer]    : Integer          read GetCellCount;
    property    Face [ aFace: Integer]                : TCubeFace        read GetFace          write SetFace;   default;
    property    Panel[ aFace, aPanel: Integer]        : TCubePanel       read GetPanel         write SetPanel;
    property    Cell [ aFace, aPanel, aCell: Integer] : TCubeCell        read GetCell          write SetCell;
    property    Pattern                               : string           read GetPattern       write SetPattern;
    property    State                                 : string           read GetState         write SetState;
    property    HistoryCount                          : Integer          read GetHistoryCount;
    property    UseRedRule                            : Boolean          read FUSeRedRule      write FUSeRedRule;
    property    Use1ForMissing                        : Boolean          read FUse1ForMissing  write SetUse1ForMissing;
    property    ShowNumbers                           : Boolean          read FShowNumbers     write SetShowNumbers;
    property    OnGetbackground                       : TOnGetbackground read FOnGetbackground write FOnGetbackground;
  end;


implementation


  function  Explode( aString: string; aSeparator: string): TStringList;
  // Note : when aSring = '' this one eturns a stringist with no
  //        items in it.
  var
    c : word;
    p : Integer;
  begin
    Result := TStringList.Create;
    c      := 0;

    while aString <> ''
    do begin
      p := Pos( aSeparator, aString);

      if   p > 0
      then begin
        Result.Add( Copy( aString, 1, p - 1));
        Delete( aString, 1, Length( Result[ c]) + Length( aSeparator));
      end
      else begin
        Result.Add( aString);
        aString := '';
      end;

      inc( c);
    end;
  end;


  procedure C( aSrc, aDst: TCubeCell);
  begin
    if   Assigned( aSrc)
    and  Assigned( aDst)
    then aSrc.ConnectTo( aDst);
  end;


  function  StrToRule( const S: string): TRule;
  // Standard Life rules:
  // S = '0,3,1; 0,4,0; 1,0,0; 1,1,0; 1,4,0; 1,5,0; 1,6,0; 1,7,0; 1,8,0'
  var
    i      : Integer;
    aParts : TStringList;
    bParts : TStringList;
    Live   : Integer;
    Count  : Integer;
    Next   : Integer;
  begin
    for i := 0 to RULE_COUNT - 1
    do begin
      Result[ False][ i] := acNothing;
      Result[ True ][ i] := acNothing;
    end;

    aParts := Explode( Trim( S), ';');

    try
      for i := 0 to aParts.Count - 1
      do begin
        bParts := Explode( Trim( aParts[ i]), ',');

        try
          if   bParts.Count = 3
          then begin
            Live  := StrToIntDef( Trim( bParts[ 0]), 0);
            Count := StrToIntDef( Trim( bParts[ 1]), 0);
            Next  := StrToIntDef( Trim( bParts[ 2]), 0);

            if   ( Next <> Live)
            and  ( Live in [ 0, 1])
            and  ( Next in [ 0, 1])
            and  ( Count >= 0)
            and  ( Count < RULE_COUNT)
            then begin
              if   Next = 0
              then Result[ True ][ Count] := acDie
              else Result[ False][ Count] := acLive;
            end;
          end;
        finally
          bParts.DisposeOf;
        end;
      end;
    finally
      aParts.DisposeOf;
    end;
  end;


{ ========
  TCell = class
  private
    FFace           : Integer;
    FPanel          : Integer;
    FCell           : Integer;
    FAlive          : Boolean;
    FNewAlive       : Boolean;
    FRed            : Boolean;
    FNewRed         : Boolean;
    FRuleGreen      : TRule;
    FRuleRed        : TRule;
    FUse1ForMissing : Boolean;
    FShowNumbers    : Boolean;
    FNeighbours     : array of TCubeCell;
  public
    property    RuleGreen                    : TRule     read GetRuleGreen      write SetRuleGreen;
    property    RuleRed                      : TRule     read GetRuleRed        write SetRuleRed;
    property    Alive                        : Boolean   read FAlive            write FAlive;
    property    Red                          : Boolean   read FRed              write FRed;
    property    State                        : Integer   read GetState          write SetState;
    property    Use1ForMissing               : Boolean   read FUse1ForMissing   write FUse1ForMissing;
    property    ShowNumbers                  : Boolean   read FShowNumbers      write FShowNumbers;
    property    Face                         : Integer   read FFace;
    property    Panel                        : Integer   read FPanel;
    property    Cell                         : Integer   read FCell;
    property    NeighbourCount               : Integer   read GetNeighbourCount;
    property    Neighbour[ anIndex: Integer] : TCubeCell read GetNeighbour;
  private
}


    function    TCubeCell.GetRuleGreen: TRule;
    begin
      Result := FRuleGreen;
    end;


    procedure   TCubeCell.SetRuleGreen( const aValue: TRule);
    begin
      FRuleGreen := aValue;
    end;


    function    TCubeCell.GetRuleRed: TRule;
    begin
      Result := FRuleRed;
    end;


    procedure   TCubeCell.SetRuleRed( const aValue: TRule);
    begin
      FRuleRed := aValue;
    end;


    function    TCubeCell.GetState: Integer;
    begin
      Result := 0;

      if   Alive
      then Result := Result + 1;

      if   Red
      then Result := Result + 2;
    end;


    procedure   TCubeCell.SetState( aValue: Integer);
    begin
      Alive := ( aValue and 1) <> 0;
      Red   := ( aValue and 2) <> 0;
    end;


    function    TCubeCell.GetNeighbourCount : Integer;
    begin
      Result := Length( FNeighbours);
    end;


    function    TCubeCell.GetNeighbour( anIndex: Integer): TCubeCell;
    begin
      if   ( anIndex >= Low ( FNeighbours))
      and  ( anIndex <= High( FNeighbours))
      then Result := FNeighbours[ anIndex]
      else Result := nil;
    end;


    function    TCubeCell.DumpNeighbours: string;
    var
      i : Integer;
    begin
      Result := '';

      for i := 0 to Length( FNeighbours) - 1
      do begin
        if   i = 0
        then Result := Format(    '[%d,%d,%d]', [         FNeighbours[ i].Face, FNeighbours[ i].Panel, FNeighbours[ i].Cell])
        else Result := Format( '%s [%d,%d,%d]', [ Result, FNeighbours[ i].Face, FNeighbours[ i].Panel, FNeighbours[ i].Cell]);
      end;
    end;


//  protected

{$ifdef USE_FMX}

    procedure   TCubeCell.PaintOn( const aCanvas: TCanvas; const aBounds: TRect; EraseBackGround: Boolean);
    const
      DELTA = 50; // 55!
    var
      aRect     : TRect;
      aTxtColor : TAlphaColor;
    begin
      aCanvas.BeginScene;

      if   EraseBackGround
      then begin
        aCanvas.Clear( claBlack);
        aCanvas.Stroke.Dash      := TStrokeDash.Solid;
        aCanvas.Stroke.Thickness := 1.0;
        aCanvas.Stroke.Color     := claGray;
        aCanvas.Stroke.Kind      := TBrushKind.Solid;
        aCanvas.Fill  .Kind      := TBrushKind.Solid;
        aCanvas.Fill  .Color     := claBlack;
        aCanvas.DrawRect( aBounds, 0, 0, AllCorners, 1.0);
      end;

      aCanvas.Stroke.Color     := $ff404040;
      aCanvas.Stroke.Thickness := 8.0;

      if   Alive
      then begin
        if   Red
        then aCanvas.Fill.Color := claYellow
        else aCanvas.Fill.Color := claLime;

        aTxtColor := claBlack;
      end
      else begin
        if   Red
        then aCanvas.Fill.Color := claRed
        else aCanvas.Fill.Color := $ff252525;

        aTxtColor := claWhite;
      end;

      aCanvas.DrawEllipse( Rect( DELTA, DELTA, aBounds.Width - DELTA, aBounds.Height - DELTA), 1.0);
      aCanvas.FillEllipse( Rect( DELTA, DELTA, aBounds.Width - DELTA, aBounds.Height - DELTA), 1.0);

      if   ShowNumbers
      then begin
        aCanvas.Fill.Color   := aTxtColor;
        aCanvas.Stroke.Color := aTxtColor;
        aCanvas.Font.Size   := 21;
        aCanvas.Font.Family := 'Arial';
        aRect.Create( 0, 0, aBounds.Width, aBounds.Height);
        aCanvas.FillText( aRect, IntToStr( FCell), False, 1.0, [ TFillTextFlag.RightToLeft], TTextAlign.Center);
      end;

      aCanvas.EndScene;
    end;

{$else}

    procedure   TCubeCell.PaintOn( const aCanvas: TCanvas; const aBounds: TRect; EraseBackGround: Boolean);
    const
      DELTA = 16;
    begin
      aCanvas.Pen  .Style := psSolid;
      aCanvas.Pen  .Color := $00252525;
      aCanvas.Brush.Style := bsSolid;
      aCanvas.Brush.Color := clBlack;
      aCanvas.Rectangle( 0, 0, aBounds.Width + 1, aBounds.Height + 1);

      aCanvas.Pen.Color := clBlack;

      if   Alive
      then begin
        if   Red
        then aCanvas.Brush.Color := clYellow
        else aCanvas.Brush.Color := clLime;
      end
      else begin
        if   Red
        then aCanvas.Brush.Color := ( clMaroon + clRed) div 2
        else aCanvas.Brush.Color := clBlack;
      end;

      aCanvas.Ellipse( DELTA, DELTA, aBounds.Width - DELTA, aBounds.Height - DELTA);
    end;

{$endif}

    function    TCubeCell.MouseToCell( anX, anY: Integer; const aBounds: TRect): TCubeCell;
    begin
      if   aBounds.Contains( Point( anX, anY))
      then Result := Self
      else Result := nil;
    end;


//  public

    constructor TCubeCell.Create( aFace, aPanel, aCell: Integer);
    begin
      FFace  := aFace;
      FPanel := aPanel;
      FCell  := aCell;
    end;


    procedure   TCubeCell.ConnectTo( const aCell: TCubeCell);
    begin
      SetLength( FNeighbours, Length( FNeighbours) + 1);
      FNeighbours[ Length( FNeighbours) - 1] := aCell;
    end;


    function    TCubeCell.CountLivingNeighbours: Integer;
    var
      i : Integer;
    begin
      Result := 0;

      for i := Low( FNeighbours) to High( FNeighbours)
      do begin
        if FNeighbours[ i].Alive
        then Inc( Result);
      end;

      if   ( NeighbourCount = 7)
      and Use1ForMissing
      then Inc( Result);
    end;


    procedure   TCubeCell.Process( UseRed: Boolean);
    var
      NC : Integer;
    begin
      NC := CountLivingNeighbours;

      case FRuleGreen[ Alive, NC] of
        acLive : FNewAlive := True;
        acDie  : FNewAlive := False;
        else     FNewAlive := Alive;
      end;

      if   UseRed
      then begin
        case FRuleRed[ Red, NC] of
          acLive : FNewRed := True;
          acDie  : FNewRed := False;
          else     FNewRed := Red;
        end;
      end
      else FNewRed := False;
    end;


    procedure   TCubeCell.Proceed;
    begin
      Alive := FNewAlive;
      Red   := FNewRed;
    end;


    function    TCubeCell.Dump: string;
    begin
      Result := Format( '[%d %d %d: %s]', [ Face, Panel, Cell, DumpNeighbours]);
    end;


    procedure   TCubeCell.Clear;
    begin
      Alive := False;
      Red   := False;
    end;


{ ========
  TPanel = class
  private
    FPArent      : TCubeFace;
    FCells       : array[ 0 .. CELLS_IN_PANEL - 1] of TCubeCell;
    FFace        : Integer;
    FPanel       : Integer;
    FShowNumbers : Boolean;
  (*$ifdef USE_FMX*)
    FBackGround  : TBitmap;
    FBackGroundR : TBitmap;
  (*$endif*)
  public
    property    RuleGreen             : TRule     read GetRuleGreen write SetRuleGreen;
    property    RuleRed               : TRule     read GetRuleRed   write SetRuleRed;
    property    Cell[ aCell: Integer] : TCubeCell read GetCell      write SetCell;                              default;
    property    State                 : string    read GetState     write SetState;
    property    Use1ForMissing        : Boolean                     write SetUse1ForMissing;
    property    ShowNumbers           : Boolean   read FShowNumbers write SetShowNumbers;
    property    Face                  : Integer   read FFace;
    property    Panel                 : Integer   read FPanel;
  private
}

    function    TCubePanel.GetCell( aCell: Integer): TCubeCell;
    begin
      if   ( aCell >= Low ( FCells))
      and  ( aCell <= High( FCells))
      then Result := FCells[ aCell]
      else Result := nil;
    end;


    procedure   TCubePanel.SetCell( aCell: Integer; const aValue: TCubeCell);
    begin
      if   ( aCell >= Low ( FCells))
      and  ( aCell <= High( FCells))
      and  Assigned( aValue)
      then FCells[ aCell].Alive := aValue.Alive;
    end;


    function    TCubePanel.GetRuleGreen: TRule;
    begin
      Result := FCells[ 0].RuleGreen;
    end;


    procedure   TCubePanel.SetRuleGreen( const aValue: TRule);
    var
      i : Integer;
    begin
      for i := Low( FCells) to High( FCells)
      do FCells[ i].RuleGreen := aValue;
    end;


    function    TCubePanel.GetRuleRed: TRule;
    begin
      Result := FCells[ 0].RuleRed;
    end;


    procedure   TCubePanel.SetRuleRed( const aValue: TRule);
    var
      i : Integer;
    begin
      for i := Low( FCells) to High( FCells)
      do FCells[ i].RuleRed := aValue;
    end;


    function    TCubePanel.GetState: string;
    begin
      Result :=
        Format(
          '%d,%d,%d,%d',
          [
            FCells[ 0].State,
            FCells[ 1].State,
            FCells[ 2].State,
            FCells[ 3].State
          ]
        );
    end;


    procedure   TCubePanel.SetState( const aValue: string);
    var
      aParts : TStringList;
      aVal   : Integer;
    begin
      aParts := Explode( Trim( aValue), ',');

      try
        if   aParts.Count = 4
        then begin
          aVal := StrToIntDef( Trim( aParts[ 0]), -1); if   aVal in [ 0 .. 3] then Cell[ 0].State := aVal;
          aVal := StrToIntDef( Trim( aParts[ 1]), -1); if   aVal in [ 0 .. 3] then Cell[ 1].State := aVal;
          aVal := StrToIntDef( Trim( aParts[ 2]), -1); if   aVal in [ 0 .. 3] then Cell[ 2].State := aVal;
          aVal := StrToIntDef( Trim( aParts[ 3]), -1); if   aVal in [ 0 .. 3] then Cell[ 3].State := aVal;
        end;
      finally
        aParts.DisposeOf;
      end;
    end;


    procedure   TCubePanel.SetUse1ForMissing( aValue: Boolean);
    var
      i : Integer;
    begin
      for i := Low( FCells) to High( FCells)
      do FCells[ i].Use1ForMissing := aValue;
    end;


    procedure   TCubePanel.SetShowNumbers( aValue: Boolean);
    var
      i : Integer;
    begin
      FShowNumbers := aValue;

      for i := Low( FCells) to High( FCells)
      do FCells[ i].ShowNumbers := aValue;
    end;


    procedure   TCubePanel.FixLocalConnections;
    begin
      C( FCells[ 0], FCells[ 1]);
      C( FCells[ 0], FCells[ 2]);
      C( FCells[ 0], FCells[ 3]);

      C( FCells[ 1], FCells[ 0]);
      C( FCells[ 1], FCells[ 2]);
      C( FCells[ 1], FCells[ 3]);

      C( FCells[ 2], FCells[ 0]);
      C( FCells[ 2], FCells[ 1]);
      C( FCells[ 2], FCells[ 3]);

      C( FCells[ 3], FCells[ 0]);
      C( FCells[ 3], FCells[ 1]);
      C( FCells[ 3], FCells[ 2]);
    end;


//  protected

{$ifdef USE_FMX}

    procedure   TCubePanel.PaintOn( const aCanvas: TCanvas; const aBounds: TRect);
    var
      PanelSize : Integer;
      aBitmap   : TBitmap;
      aRect     : TRect;
      aFileName : string;
      anErase   : Boolean;
    begin
      PanelSize := aBounds.Width div Round( Sqrt( CELLS_IN_PANEL));
      aBitmap   := TBitmap.Create;

      try
        aBitmap.Width  := PanelSize;
        aBitmap.Height := PanelSize;
        anErase        := True;

        if   not Assigned( FBackGround)
        then begin
          aFileName := Getbackground;

          if   FileExists( aFileName)
          then begin
            FBackGround  := TBitmap.Create;
            FBackGround.LoadFromFile( aFileName);
            FBackGroundR := TBitmap.Create;
            FBackGroundR.LoadFromFile( aFileName);
            FBackGroundR.Rotate( 90);
          end;
        end
        else begin
          aFileName := Getbackground;

          if   not FileExists( aFileName)
          then begin
            FreeAndNil( FBackGround );
            FreeAndNil( FBackGroundR);
          end;
        end;

        if   Assigned( FBackGround)
        then begin
          aCanvas.BeginScene;

          try
            if   FFace in [ 2, 4] // Draw totated bitmaps for faces 2 and 4
            then aCanvas.DrawBitmap( FBackGroundR, RectF( 0, 0, FBackGroundR.Width, FBackGroundR.Height), aBounds, 1.0)
            else aCanvas.DrawBitmap( FBackGround , RectF( 0, 0, FBackGround .Width, FBackGround .Height), aBounds, 1.0);
            anErase := False;
          finally
            aCanvas.EndScene;
          end;
        end;

        aCanvas.BeginScene;
        Cell[ 0].PaintOn( aBitmap.Canvas, Rect( 1, 1, PanelSize, PanelSize), anErase); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 0.0, PanelSize * 0.0, PanelSize * 1.0, PanelSize * 1.0), 1.0);
        aCanvas.EndScene;
        aCanvas.BeginScene;
        Cell[ 1].PaintOn( aBitmap.Canvas, Rect( 1, 1, PanelSize, PanelSize), anErase); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 1.0, PanelSize * 0.0, PanelSize * 2.0, PanelSize * 1.0), 1.0);
        aCanvas.EndScene;
        aCanvas.BeginScene;
        Cell[ 2].PaintOn( aBitmap.Canvas, Rect( 1, 1, PanelSize, PanelSize), anErase); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 0.0, PanelSize * 1.0, PanelSize * 1.0, PanelSize * 2.0), 1.0);
        aCanvas.EndScene;
        aCanvas.BeginScene;
        Cell[ 3].PaintOn( aBitmap.Canvas, Rect( 1, 1, PanelSize, PanelSize), anErase); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 1.0, PanelSize * 1.0, PanelSize * 2.0, PanelSize * 2.0), 1.0);

        if   ShowNumbers
        then begin
          aCanvas.Font.Size   := 27;
          aCanvas.Font.Family := 'Arial';
          aRect.Create( 0, 0, aBounds.Width, aBounds.Height);
          aCanvas.FillText( aRect, IntToStr( FPanel), False, 1.0, [ TFillTextFlag.RightToLeft], TTextAlign.Center);
        end;

        aCanvas.EndScene;
      finally
        aBitmap.DisposeOf;
      end;
    end;

{$else}

    procedure   TCubePanel.PaintOn( const aCanvas: TCanvas; const aBounds: TRect);
    var
      PanelSize : Integer;
      aBitmap   : TBitmap;
    begin
      PanelSize := aBounds.Width div Round( Sqrt( CELLS_IN_PANEL));
      aBitmap   := TBitmap.Create;

      try
        aBitmap.Width  := PanelSize;
        aBitmap.Height := PanelSize;
        Cell[ 0].PaintOn( aBitmap.Canvas, Rect( 1, 1, PanelSize, PanelSize), True); aCanvas.Draw( 0        , 0        , aBitmap);
        Cell[ 1].PaintOn( aBitmap.Canvas, Rect( 1, 1, PanelSize, PanelSize), True); aCanvas.Draw( PanelSize, 0        , aBitmap);
        Cell[ 2].PaintOn( aBitmap.Canvas, Rect( 1, 1, PanelSize, PanelSize), True); aCanvas.Draw( 0        , PanelSize, aBitmap);
        Cell[ 3].PaintOn( aBitmap.Canvas, Rect( 1, 1, PanelSize, PanelSize), True); aCanvas.Draw( PanelSize, PanelSize, aBitmap);
      finally
        aBitmap.DisposeOf;
      end;
    end;

{$endif}

    function    TCubePanel.MouseToCell( anX, anY: Integer; const aBounds: TRect): TCubeCell;
    var
      PanelSize : Integer;
      R         : TRect;
    begin
      Result := nil;

      if   aBounds.Contains( Point( anX, anY))
      then begin
        PanelSize := aBounds.Width div Round( Sqrt( PANELS_IN_FACE));
        R         := Rect( 0, 0, PanelSize, PanelSize);
        Result    := Cell[ 0].MouseToCell( anX            , anY            , R); if Assigned( Result) then Exit;
        Result    := Cell[ 1].MouseToCell( anX - PanelSize, anY            , R); if Assigned( Result) then Exit;
        Result    := Cell[ 2].MouseToCell( anX            , anY - PanelSize, R); if Assigned( Result) then Exit;
        Result    := Cell[ 3].MouseToCell( anX - PanelSize, anY - PanelSize, R); if Assigned( Result) then Exit;
      end
    end;


    function    TCubePanel.Getbackground: string;
    begin
      if   Assigned( FParent)
      then Result := FParent.Getbackground
      else Result := '';
    end;


//  public

    constructor TCubePanel.Create( const aParent: TCubeFace; aFace, aPanel: Integer);
    var
      i : Integer;
    begin
      FParent := aParent;
      FFace   := aFace;
      FPanel  := aPanel;

      for i := Low( FCells) to High( FCells)
      do begin
        FCells[ i] := TCubeCell.Create( aFace, aPanel, i);
        FCells[ i].Alive := False;
      end;

      FixLocalConnections;
    end;


    destructor  TCubePanel.Destroy; // override;
    var
      i : Integer;
    begin
      for i := Low( FCells) to High( FCells)
      do FreeAndNil( FCells[ i]);

      FreeAndNil( FBackGround );
      FreeAndNil( FBackGroundR);
      inherited;
    end;


    procedure   TCubePanel.Process( UseRed: Boolean);
    var
      i : Integer;
    begin
      for i := Low( FCells) to High( FCells)
      do FCells[ i].Process( UseRed);
    end;


    procedure   TCubePanel.Proceed;
    var
      i : Integer;
    begin
      for i := Low( FCells) to High( FCells)
      do FCells[ i].Proceed;
    end;


    function    TCubePanel.Dump: string;
    var
      i : Integer;
    begin
      Result := '';

      for i := Low( FCells) to High( FCells)
      do Result := Result + Format( '%s'^M, [ FCells[ i].Dump]);
    end;


    procedure   TCubePanel.Clear;
    var
      i : Integer;
    begin
      for i := Low( FCells) to High( FCells)
      do FCells[ i].Clear;
    end;


{ ========
  TFace = class
  private
    FPanels      : array[ 0 .. PANELS_IN_FACE - 1] of TCubePanel;
    FFace        : Integer;
    FShowNumbers : Boolean;
  public
    property    RuleGreen                      : TRule      read GetRuleGreen  write SetRuleGreen;
    property    RuleRed                        : TRule      read GetRuleRed    write SetRuleRed;
    property    Panel[ aPanel: Integer       ] : TCubePanel read GetPanel      write SetPanel;                  default;
    property    Cell [ aPanel, aCell: Integer] : TCubeCell  read GetCell       write SetCell;
    property    State                          : string     read GetState      write SetState;
    property    Use1ForMissing                 : Boolean                       write SetUse1ForMissing;
    property    ShowNumbers                    : Boolean    read FShowNumbers  write SetShowNumbers;
    property    Face                           : Integer    read FFace;
  private
}

    function    TCubeFace.GetPanel( aPanel: Integer): TCubePanel;
    begin
      if   ( aPanel >= Low ( FPanels))
      and  ( aPanel <= High( FPanels))
      then Result := FPanels[ aPanel]
      else Result := nil;
    end;


    procedure   TCubeFace.SetPanel( aPanel: Integer; const aValue: TCubePanel);
    var
      i : Integer;
    begin
      if   ( aPanel >= Low ( FPanels))
      and  ( aPanel <= High( FPanels))
      and  Assigned( aValue)
      then begin
        for i := 0 to 3
        do FPanels[ aPanel].Cell[ i] := aValue.Cell[ i];
      end;
    end;


    function    TCubeFace.GetCell( aPanel, aCell: Integer): TCubeCell;
    begin
      Result := FPanels[ aPanel][ aCell];
    end;


    procedure   TCubeFace.SetCell( aPanel, aCell: Integer; const aValue: TCubeCell);
    begin
      FPanels[ aPanel][ aCell] := aValue;
    end;


    function    TCubeFace.GetRuleGreen: TRule;
    begin
      Result := FPanels[ 0].RuleGreen;
    end;


    procedure   TCubeFace.SetRuleGreen( const aValue: TRule);
    var
      i : Integer;
    begin
      for i := Low( FPanels) to High( FPanels)
      do FPanels[ i].RuleGreen := aValue;
    end;


    function    TCubeFace.GetRuleRed: TRule;
    begin
      Result := FPanels[ 0].RuleRed;
    end;


    procedure   TCubeFace.SetRuleRed( const aValue: TRule);
    var
      i : Integer;
    begin
      for i := Low( FPanels) to High( FPanels)
      do FPanels[ i].RuleRed := aValue;
    end;


    function    TCubeFace.GetState: string;
    begin
      Result :=
        Format(
          '%s;%s;%s;%s',
          [
            FPanels[ 0].State,
            FPanels[ 1].State,
            FPanels[ 2].State,
            FPanels[ 3].State
          ]
        );
    end;


    procedure   TCubeFace.SetState( const aValue: string);
    var
      aParts : TStringList;
    begin
      aParts := Explode( Trim( aValue), ';');

      try
        if   aParts.Count = 4
        then begin
          FPanels[ 0].State := Trim( aParts[ 0]);
          FPanels[ 1].State := Trim( aParts[ 1]);
          FPanels[ 2].State := Trim( aParts[ 2]);
          FPanels[ 3].State := Trim( aParts[ 3]);
        end;
      finally
        aParts.DisposeOf;
      end;
    end;


    procedure   TCubeFace.SetUse1ForMissing( aValue: Boolean);
    var
      i : Integer;
    begin
      for i := Low( FPanels) to High( FPanels)
      do FPanels[ i].Use1ForMissing := aValue;
    end;


    procedure   TCubeFace.SetShowNumbers( aValue: Boolean);
    var
      i : Integer;
    begin
      FShowNumbers := aValue;

      for i := Low( FPanels) to High( FPanels)
      do FPanels[ i].ShowNumbers := aValue;
    end;


    procedure   TCubeFace.FixLocalConnections;
    begin
      C( FPanels[ 0][ 1], FPanels[ 1][ 0]);
      C( FPanels[ 0][ 1], FPanels[ 1][ 2]);

      C( FPanels[ 0][ 2], FPanels[ 2][ 0]);
      C( FPanels[ 0][ 2], FPanels[ 2][ 1]);

      C( FPanels[ 0][ 3], FPanels[ 1][ 0]);
      C( FPanels[ 0][ 3], FPanels[ 1][ 2]);
      C( FPanels[ 0][ 3], FPanels[ 2][ 0]);
      C( FPanels[ 0][ 3], FPanels[ 2][ 1]);
      C( FPanels[ 0][ 3], FPanels[ 3][ 0]);

      // ----------------------------------

      C( FPanels[ 1][ 0], FPanels[ 0][ 1]);
      C( FPanels[ 1][ 0], FPanels[ 0][ 3]);

      C( FPanels[ 1][ 2], FPanels[ 0][ 1]);
      C( FPanels[ 1][ 2], FPanels[ 0][ 3]);
      C( FPanels[ 1][ 2], FPanels[ 2][ 1]);
      C( FPanels[ 1][ 2], FPanels[ 3][ 0]);
      C( FPanels[ 1][ 2], FPanels[ 3][ 1]);

      C( FPanels[ 1][ 3], FPanels[ 3][ 0]);
      C( FPanels[ 1][ 3], FPanels[ 3][ 1]);

      // ----------------------------------

      C( FPanels[ 2][ 0], FPanels[ 0][ 2]);
      C( FPanels[ 2][ 0], FPanels[ 0][ 3]);

      C( FPanels[ 2][ 1], FPanels[ 0][ 2]);
      C( FPanels[ 2][ 1], FPanels[ 0][ 3]);
      C( FPanels[ 2][ 1], FPanels[ 1][ 2]);
      C( FPanels[ 2][ 1], FPanels[ 3][ 0]);
      C( FPanels[ 2][ 1], FPanels[ 3][ 2]);

      C( FPanels[ 2][ 3], FPanels[ 3][ 0]);
      C( FPanels[ 2][ 3], FPanels[ 3][ 2]);

      // ----------------------------------

      C( FPanels[ 3][ 0], FPanels[ 0][ 3]);
      C( FPanels[ 3][ 0], FPanels[ 1][ 2]);
      C( FPanels[ 3][ 0], FPanels[ 1][ 3]);
      C( FPanels[ 3][ 0], FPanels[ 2][ 1]);
      C( FPanels[ 3][ 0], FPanels[ 2][ 3]);

      C( FPanels[ 3][ 1], FPanels[ 1][ 2]);
      C( FPanels[ 3][ 1], FPanels[ 1][ 3]);

      C( FPanels[ 3][ 2], FPanels[ 2][ 1]);
      C( FPanels[ 3][ 2], FPanels[ 2][ 3]);
    end;


//  protected

{$ifdef USE_FMX}

    procedure   TCubeFace.PaintOn( const aCanvas: TCanvas; const aBounds: TRect);
    var
      PanelSize : Integer;
      aBitmap   : TBitmap;
      aRect     : TRect;
    begin
      PanelSize := aBounds.Width div 2;
      aBitmap   := TBitmap.Create;

      try
        aBitmap.Width  := PanelSize;
        aBitmap.Height := PanelSize;
        aCanvas.BeginScene;
        FPanels[ 0].PaintOn( aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 0.0, PanelSize * 0.0, PanelSize * 1.0, PanelSize * 1.0), 1.0);
        aCanvas.EndScene;
        aCanvas.BeginScene;
        FPanels[ 1].PaintOn( aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 1.0, PanelSize * 0.0, PanelSize * 2.0, PanelSize * 1.0), 1.0);
        aCanvas.EndScene;
        aCanvas.BeginScene;
        FPanels[ 2].PaintOn( aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 0.0, PanelSize * 1.0, PanelSize * 1.0, PanelSize * 2.0), 1.0);
        aCanvas.EndScene;
        aCanvas.BeginScene;
        FPanels[ 3].PaintOn( aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 1.0, PanelSize * 1.0, PanelSize * 2.0, PanelSize * 2.0), 1.0);

        if   ShowNumbers
        then begin
          aCanvas.Font.Size   := 33;
          aCanvas.Font.Family := 'Arial';
          aRect.Create( 0, 0, aBounds.Width, aBounds.Height);
          aCanvas.FillText( aRect, IntToStr( FFace), False, 1.0, [ TFillTextFlag.RightToLeft], TTextAlign.Center);
        end;

        aCanvas.EndScene;
      finally
        aBitmap.DisposeOf;
      end;
    end;

{$else}

    procedure   TCubeFace.PaintOn( const aCanvas: TCanvas; const aBounds: TRect);
    var
      PanelSize : Integer;
      aBitmap   : TBitmap;
    begin
      PanelSize := aBounds.Width div 2;
      aBitmap   := TBitmap.Create;

      try
        aBitmap.Width  := PanelSize;
        aBitmap.Height := PanelSize;
        FPanels[ 0].PaintOn( aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.Draw( 0        , 0        , aBitmap);
        FPanels[ 1].PaintOn( aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.Draw( PanelSize, 0        , aBitmap);
        FPanels[ 2].PaintOn( aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.Draw( 0        , PanelSize, aBitmap);
        FPanels[ 3].PaintOn( aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.Draw( PanelSize, PanelSize, aBitmap);
      finally
        aBitmap.DisposeOf;
      end;
    end;

{$endif}

    function    TCubeFace.MouseToCell( anX, anY: Integer; const aBounds: TRect): TCubeCell;
    var
      PanelSize : Integer;
      R         : TRect;
    begin
      Result := nil;

      if   aBounds.Contains( Point( anX, anY))
      then begin
        PanelSize := aBounds.Width div 2;
        R         := Rect( 0, 0, PanelSize, PanelSize);
        Result    := FPanels[ 0].MouseToCell( anX            , anY            , R); if Assigned( Result) then Exit;
        Result    := FPanels[ 1].MouseToCell( anX - PanelSize, anY            , R); if Assigned( Result) then Exit;
        Result    := FPanels[ 2].MouseToCell( anX            , anY - PanelSize, R); if Assigned( Result) then Exit;
        Result    := FPanels[ 3].MouseToCell( anX - PanelSize, anY - PanelSize, R); if Assigned( Result) then Exit;
      end;
    end;


    function    TCubeFace.Getbackground: string;
    begin
      if   Assigned( FParent)
      then Result := FParent.Getbackground
      else Result := '';
    end;


//  public

    constructor TCubeFace.Create( const aParent: TCubeLife; aFace: Integer);
    var
      i : Integer;
    begin
      FParent := aParent;
      FFace   := aFace;

      for i := Low( FPanels) to High( FPanels)
      do FPanels[ i] := TCubePanel.Create( Self, aFace, i);

      FixLocalConnections;
    end;


    destructor  TCubeFace.Destroy; // override;
    var
      i : Integer;
    begin
      for i := Low( FPanels) to High( FPanels)
      do FreeAndNil( FPanels[ i]);

      inherited;
    end;


    procedure   TCubeFace.Process( UseRed: Boolean);
    var
      i : Integer;
    begin
      for i := Low( FPanels) to High( FPanels)
      do FPanels[ i].Process( UseRed);
    end;


    procedure   TCubeFace.Proceed;
    var
      i : Integer;
    begin
      for i := Low( FPanels) to High( FPanels)
      do FPanels[ i].Proceed;
    end;


    function    TCubeFace.Dump: string;
    var
      i : Integer;
    begin
      Result := '';

      for i := Low( FPanels) to High( FPanels)
      do Result := Result + Format( '%s'^M, [ FPanels[ i].Dump]);
    end;


    procedure   TCubeFace.Clear;
    var
      i : Integer;
    begin
      for i := Low( FPanels) to High( FPanels)
      do FPanels[ i].Clear;
    end;


{ ========
  TCubeHistory = class
  private
    FData       : array of string;
    FFillCount  : Integer;
    FFreeCount  : Integer;
    FInPointer  : Integer;
    FOutPointer : Integer;
  public
    property    FillCount: Integer read FFillCount;
  public
}

    constructor TCubeHistory.Create( aSize: Integer);
    begin
      inherited Create;
      SetLength( FData, aSize);
      Clear;
    end;


    destructor  TCubeHistory.Destroy; // override;
    begin
      SetLength( FData, 0);
      inherited;
    end;


    procedure   TCubeHistory.Clear;
    begin
      FFillCount := 0;
      FFreeCount := 0;
      FPointer   := 0;
      FFreeCount := Length( FData);
    end;


    procedure   TCubeHistory.Push( const aValue: string);
    begin
      if   FFreeCount <= 0
      then begin
        FPointer := ( FPointer + 1) mod Length( FData);
        FData[ FPointer] := aValue;
      end
      else begin
        Dec( FFreeCount);
        FData[ FPointer] := aValue;
        FPointer := ( FPointer + 1) mod Length( FData);
        Inc( FFillCount);
      end;
    end;


    function    TCubeHistory.Pop: string;
    begin
      if   FillCount > 0
      then begin
        Dec( FFillCount);

        if   FPointer > 0
        then Dec( FPointer)
        else FPointer := Length( FData) - 1;

        Result := FData[ FPointer];
        Inc( FFreeCount);
      end
      else Result := '';
    end;


{ ========
  TCubeLife = class
  private
    FFaces           : array[ 0 .. FACES_IN_CUBE - 1] of TCubeFace;
    FRuleStrGreen    : string;
    FRuleStrRed      : string;
    FHistory         : TCubeHistory;
    FUseRedRule      : Boolean;
    FUse1ForMissing  : Boolean;
    FShowNumbers     : Boolean;
    FOnGetBackground : TOnGetBackground;
  public
    property    RuleGreen                             : TRule            read GetRuleGreen     write SetRuleGreen;
    property    RuleRed                               : TRule            read GetRuleRed       write SetRuleRed;
    property    RuleStrGreen                          : String           read FRuleStrGreen    write SetRuleStrGreen;
    property    RuleStrRed                            : String           read FRuleStrRed      write SetRuleStrRed;
    property    FaceCount                             : Integer          read GetFaceCount;
    property    PanelCount[ aFace: Integer]           : Integer          read GetPanelCount;
    property    CellCount[ aFace, aPanel: Integer]    : Integer          read GetCellCount;
    property    Face [ aFace: Integer]                : TCubeFace        read GetFace          write SetFace;   default;
    property    Panel[ aFace, aPanel: Integer]        : TCubePanel       read GetPanel         write SetPanel;
    property    Cell [ aFace, aPanel, aCell: Integer] : TCubeCell        read GetCell          write SetCell;
    property    Pattern                               : string           read GetPattern       write SetPattern;
    property    State                                 : string           read GetState         write SetState;
    property    HistoryCount                          : Integer          read GetHistoryCount;
    property    UseRedRule                            : Boolean          read FUSeRedRule      write FUSeRedRule;
    property    Use1ForMissing                        : Boolean          read FUse1ForMissing  write SetUse1ForMissing;
    property    ShowNumbers                           : Boolean          read FShowNumbers     write SetShowNumbers;
    property    OnGetbackground                       : TOnGetbackground read FOnGetbackground write FOnGetbackground;
  private
}

    function    TCubeLife.GetFace( aFace: Integer): TCubeFace;
    begin
      if   ( aFace >= Low ( FFaces))
      and  ( aFace <= High( FFaces))
      then Result := FFaces[ aFace]
      else Result := nil;
    end;


    procedure   TCubeLife.SetFace( aFace: Integer; const aValue: TCubeFace);
    var
      i : Integer;
    begin
      if   ( aFace >= Low ( FFaces))
      and  ( aFace <= High( FFaces))
      and  Assigned( aValue)
      then begin
        for i := 0 to 3
        do FFaces[ aFace].Panel[ i] := aValue.Panel[ i];
      end;
    end;


    function    TCubeLife.GetPanel( aFace, aPanel: Integer): TCubePanel;
    begin
      Result := Face[ aFace][ aPanel];
    end;


    procedure   TCubeLife.SetPanel( aFace, aPanel: Integer; const aValue: TCubePanel);
    begin
      Face[ aFace][ aPanel] := aValue;
    end;


    function    TCubeLife.GetCell( aFace, aPanel, aCell: Integer): TCubeCell;
    begin
      Result := Panel[ aFace, aPanel][ aCell];
    end;


    procedure   TCubeLife.SetCell( aFace, aPanel, aCell: Integer; const aValue: TCubeCell);
    begin
      Panel[ aFace, aPanel][ aCell] := aValue;
    end;


    function    TCubeLife.GetFaceCount: Integer;
    begin
      Result := FACES_IN_CUBE;
    end;


    function    TCubeLife.GetPanelCount( aFace: Integer): Integer;
    begin
      Result := PANELS_IN_FACE;
    end;


    function    TCubeLife.GetCellCount ( aFace, aPanel: Integer): Integer;
    begin
      Result := CELLS_IN_PANEL;
    end;


    function    TCubeLife.GetRuleGreen: TRule;
    begin
      Result := FFaces[ 0].RuleGreen;
    end;


    procedure   TCubeLife.SetRuleGreen( const aValue: TRule);
    var
      i : Integer;
    begin
      for i := Low( FFaces) to High( FFaces)
      do FFaces[ i].RuleGreen := aValue;
    end;


    function    TCubeLife.GetRuleRed: TRule;
    begin
      Result := FFaces[ 0].RuleRed;
    end;


    procedure   TCubeLife.SetRuleRed( const aValue: TRule);
    var
      i : Integer;
    begin
      for i := Low( FFaces) to High( FFaces)
      do FFaces[ i].RuleRed := aValue;
    end;


    procedure   TCubeLife.SetRuleStrGreen( const aValue: string);
    begin
      FRuleStrGreen := aValue;
      RuleGreen     := StrToRule( aValue);
    end;


    procedure   TCubeLife.SetRuleStrRed( const aValue: string);
    begin
      FRuleStrRed := aValue;
      RuleRed     := StrToRule( aValue);
    end;


    function    TCubeLife.GetPattern: string;
    var
      aFace  : Integer;
      aPanel : Integer;
      aCell  : Integer;
      C      : TCubeCell;
      S      : string;
    begin
      Result := '';

      for aFace := 0 to FaceCount - 1
      do begin
        for aPanel := 0 to PanelCount[ aFace] - 1
        do begin
          for aCell := 0 to CellCount[ aFace, aPanel] - 1
          do begin
            C := Cell[ aFace, aPanel, aCell];

            if   C.Alive
            then begin
              S := Format( '%d,%d,%d', [ aFace, aPanel, aCell]);

              if   Result = ''
              then Result := S
              else Result := Result + ';' + S;
            end;
          end;
        end;
      end;
    end;


    procedure   TCubeLife.SetPattern( const aValue: string);
    var
      aParts : TStringList;
      bParts : TStringList;
      i      : Integer;
      aFace  : Integer;
      aPanel : Integer;
      aCell  : Integer;
    begin
      Clear;
      aParts := Explode( Trim( aValue), ';');

      try
        for i := 0 to aParts.Count - 1
        do begin
          bParts := Explode( Trim( aParts[ i]), ',');

          try
            if   bParts.Count = 3
            then begin
              aFace  := StrToIntDef( Trim( bParts[ 0]), -1);
              aPanel := StrToIntDef( Trim( bParts[ 1]), -1);
              aCell  := StrToIntDef( Trim( bParts[ 2]), -1);

              if   ( aFace  >= 0                        )
              and  ( aFace  <  FaceCount                )
              and  ( aPanel >= 0                        )
              and  ( aPanel <  PanelCount[ aFace]       )
              and  ( aCell  >= 0                        )
              and  ( aCell  <  CellCount[ aFace, aPanel])
              then Cell[ aFace, aPanel, aCell].Alive := True;
            end;
          finally
            bParts.DisposeOf;
          end;
        end;
      finally
        aParts.DisposeOf;
      end;
    end;


    function    TCubeLife.GetState: string;
    begin
      Result :=
        Format(
          '%s:%s:%s:%s:%s:%s',
          [
            FFaces[ 0].State,
            FFaces[ 1].State,
            FFaces[ 2].State,
            FFaces[ 3].State,
            FFaces[ 4].State,
            FFaces[ 5].State
          ]
        );
    end;


    procedure   TCubeLife.SetState( const aValue: string);
    var
      aParts : TStringList;
    begin
      aParts := Explode( Trim( aValue), ':');

      try
        if   aParts.Count = 6
        then begin
          FFaces[ 0].State := Trim( aParts[ 0]);
          FFaces[ 1].State := Trim( aParts[ 1]);
          FFaces[ 2].State := Trim( aParts[ 2]);
          FFaces[ 3].State := Trim( aParts[ 3]);
          FFaces[ 4].State := Trim( aParts[ 2]);
          FFaces[ 5].State := Trim( aParts[ 3]);
        end;
      finally
        aParts.DisposeOf;
      end;
    end;


    function    TCubeLife.GetHistoryCount : Integer;
    begin
      if   Assigned( FHistory)
      then Result := FHistory.FillCount
      else Result := 0;
    end;


    procedure   TCubeLife.SetUse1ForMissing( aValue: Boolean);
    var
      i : Integer;
    begin
      if   aValue <> FUse1ForMissing
      then begin
        FUse1ForMissing := aValue;

        for i := Low( FFaces) to High( FFaces)
        do FFaces[ i].Use1ForMissing := aValue;
      end;
    end;


    procedure   TCubeLife.SetShowNumbers( aValue: Boolean);
    var
      i : Integer;
    begin
      if   aValue <> FShowNumbers
      then begin
        FShowNumbers := aValue;

        for i := Low( FFaces) to High( FFaces)
        do FFaces[ i].ShowNumbers := aValue;
      end;
    end;


    procedure   TCubeLife.FixLocalConnections;
    begin
      // ---- Face 0 --------------------

      C( Cell[ 0, 0, 0], Cell[ 1, 0, 0]);
      C( Cell[ 0, 0, 0], Cell[ 1, 0, 1]);
      C( Cell[ 0, 0, 0], Cell[ 4, 1, 0]);
      C( Cell[ 0, 0, 0], Cell[ 4, 1, 1]);

      C( Cell[ 0, 0, 1], Cell[ 4, 0, 1]);
      C( Cell[ 0, 0, 1], Cell[ 4, 1, 0]);
      C( Cell[ 0, 0, 1], Cell[ 4, 1, 1]);

      C( Cell[ 0, 0, 2], Cell[ 1, 0, 0]);
      C( Cell[ 0, 0, 2], Cell[ 1, 0, 1]);
      C( Cell[ 0, 0, 2], Cell[ 1, 1, 0]);

      // --------------------------------

      C( Cell[ 0, 1, 0], Cell[ 4, 0, 0]);
      C( Cell[ 0, 1, 0], Cell[ 4, 0, 1]);
      C( Cell[ 0, 1, 0], Cell[ 4, 1, 0]);

      C( Cell[ 0, 1, 1], Cell[ 4, 0, 0]);
      C( Cell[ 0, 1, 1], Cell[ 4, 0, 1]);
      C( Cell[ 0, 1, 1], Cell[ 3, 1, 0]);
      C( Cell[ 0, 1, 1], Cell[ 3, 1, 1]);

      C( Cell[ 0, 1, 3], Cell[ 3, 0, 1]);
      C( Cell[ 0, 1, 3], Cell[ 3, 1, 0]);
      C( Cell[ 0, 1, 3], Cell[ 3, 1, 1]);

      // --------------------------------

      C( Cell[ 0, 2, 0], Cell[ 1, 0, 1]);
      C( Cell[ 0, 2, 0], Cell[ 1, 1, 0]);
      C( Cell[ 0, 2, 0], Cell[ 1, 1, 1]);

      C( Cell[ 0, 2, 2], Cell[ 1, 1, 0]);
      C( Cell[ 0, 2, 2], Cell[ 1, 1, 1]);
      C( Cell[ 0, 2, 2], Cell[ 2, 0, 0]);
      C( Cell[ 0, 2, 2], Cell[ 2, 0, 1]);

      C( Cell[ 0, 2, 3], Cell[ 2, 0, 0]);
      C( Cell[ 0, 2, 3], Cell[ 2, 0, 1]);
      C( Cell[ 0, 2, 3], Cell[ 2, 1, 0]);

      // --------------------------------

      C( Cell[ 0, 3, 1], Cell[ 3, 0, 0]);
      C( Cell[ 0, 3, 1], Cell[ 3, 0, 1]);
      C( Cell[ 0, 3, 1], Cell[ 3, 1, 0]);

      C( Cell[ 0, 3, 2], Cell[ 2, 0, 1]);
      C( Cell[ 0, 3, 2], Cell[ 2, 1, 0]);
      C( Cell[ 0, 3, 2], Cell[ 2, 1, 1]);

      C( Cell[ 0, 3, 3], Cell[ 2, 1, 0]);
      C( Cell[ 0, 3, 3], Cell[ 2, 1, 1]);
      C( Cell[ 0, 3, 3], Cell[ 3, 0, 0]);
      C( Cell[ 0, 3, 3], Cell[ 3, 0, 1]);

      // ---- Face 1 --------------------

      C( Cell[ 1, 0, 0], Cell[ 0, 0, 0]);
      C( Cell[ 1, 0, 0], Cell[ 0, 0, 2]);
      C( Cell[ 1, 0, 0], Cell[ 4, 1, 1]);
      C( Cell[ 1, 0, 0], Cell[ 4, 1, 3]);

      C( Cell[ 1, 0, 1], Cell[ 0, 0, 0]);
      C( Cell[ 1, 0, 1], Cell[ 0, 0, 2]);
      C( Cell[ 1, 0, 1], Cell[ 0, 2, 0]);

      C( Cell[ 1, 0, 2], Cell[ 4, 1, 1]);
      C( Cell[ 1, 0, 2], Cell[ 4, 1, 3]);
      C( Cell[ 1, 0, 2], Cell[ 4, 3, 1]);

      // --------------------------------

      C( Cell[ 1, 1, 0], Cell[ 0, 0, 2]);
      C( Cell[ 1, 1, 0], Cell[ 0, 2, 0]);
      C( Cell[ 1, 1, 0], Cell[ 0, 2, 2]);

      C( Cell[ 1, 1, 1], Cell[ 0, 2, 0]);
      C( Cell[ 1, 1, 1], Cell[ 0, 2, 2]);
      C( Cell[ 1, 1, 1], Cell[ 2, 0, 0]);
      C( Cell[ 1, 1, 1], Cell[ 2, 0, 2]);

      C( Cell[ 1, 1, 3], Cell[ 2, 0, 0]);
      C( Cell[ 1, 1, 3], Cell[ 2, 0, 2]);
      C( Cell[ 1, 1, 3], Cell[ 2, 2, 0]);

      // --------------------------------

      C( Cell[ 1, 2, 0], Cell[ 4, 1, 3]);
      C( Cell[ 1, 2, 0], Cell[ 4, 3, 1]);
      C( Cell[ 1, 2, 0], Cell[ 4, 3, 3]);

      C( Cell[ 1, 2, 2], Cell[ 4, 3, 1]);
      C( Cell[ 1, 2, 2], Cell[ 4, 3, 3]);
      C( Cell[ 1, 2, 2], Cell[ 5, 2, 0]);
      C( Cell[ 1, 2, 2], Cell[ 5, 2, 2]);

      C( Cell[ 1, 2, 3], Cell[ 5, 0, 2]);
      C( Cell[ 1, 2, 3], Cell[ 5, 2, 0]);
      C( Cell[ 1, 2, 3], Cell[ 5, 2, 2]);

      // --------------------------------

      C( Cell[ 1, 3, 1], Cell[ 2, 0, 2]);
      C( Cell[ 1, 3, 1], Cell[ 2, 2, 0]);
      C( Cell[ 1, 3, 1], Cell[ 2, 2, 2]);

      C( Cell[ 1, 3, 2], Cell[ 5, 0, 0]);
      C( Cell[ 1, 3, 2], Cell[ 5, 0, 2]);
      C( Cell[ 1, 3, 2], Cell[ 5, 2, 0]);

      C( Cell[ 1, 3, 3], Cell[ 2, 2, 0]);
      C( Cell[ 1, 3, 3], Cell[ 2, 2, 2]);
      C( Cell[ 1, 3, 3], Cell[ 5, 0, 0]);
      C( Cell[ 1, 3, 3], Cell[ 5, 0, 2]);

      // ---- Face 2 --------------------

      C( Cell[ 2, 0, 0], Cell[ 0, 2, 2]);
      C( Cell[ 2, 0, 0], Cell[ 0, 2, 3]);
      C( Cell[ 2, 0, 0], Cell[ 1, 1, 1]);
      C( Cell[ 2, 0, 0], Cell[ 1, 1, 3]);

      C( Cell[ 2, 0, 1], Cell[ 0, 2, 2]);
      C( Cell[ 2, 0, 1], Cell[ 0, 2, 3]);
      C( Cell[ 2, 0, 1], Cell[ 0, 3, 2]);

      C( Cell[ 2, 0, 2], Cell[ 1, 1, 1]);
      C( Cell[ 2, 0, 2], Cell[ 1, 1, 3]);
      C( Cell[ 2, 0, 2], Cell[ 1, 3, 1]);

      // --------------------------------

      C( Cell[ 2, 1, 0], Cell[ 0, 2, 3]);
      C( Cell[ 2, 1, 0], Cell[ 0, 3, 2]);
      C( Cell[ 2, 1, 0], Cell[ 0, 3, 3]);

      C( Cell[ 2, 1, 1], Cell[ 0, 3, 2]);
      C( Cell[ 2, 1, 1], Cell[ 0, 3, 3]);
      C( Cell[ 2, 1, 1], Cell[ 3, 0, 0]);
      C( Cell[ 2, 1, 1], Cell[ 3, 0, 2]);

      C( Cell[ 2, 1, 3], Cell[ 3, 0, 0]);
      C( Cell[ 2, 1, 3], Cell[ 3, 0, 2]);
      C( Cell[ 2, 1, 3], Cell[ 3, 2, 0]);

      // --------------------------------

      C( Cell[ 2, 2, 0], Cell[ 1, 1, 3]);
      C( Cell[ 2, 2, 0], Cell[ 1, 3, 1]);
      C( Cell[ 2, 2, 0], Cell[ 1, 3, 3]);

      C( Cell[ 2, 2, 2], Cell[ 1, 3, 1]);
      C( Cell[ 2, 2, 2], Cell[ 1, 3, 3]);
      C( Cell[ 2, 2, 2], Cell[ 5, 0, 0]);
      C( Cell[ 2, 2, 2], Cell[ 5, 0, 1]);

      C( Cell[ 2, 2, 3], Cell[ 5, 0, 0]);
      C( Cell[ 2, 2, 3], Cell[ 5, 0, 1]);
      C( Cell[ 2, 2, 3], Cell[ 5, 1, 0]);

      // --------------------------------

      C( Cell[ 2, 3, 1], Cell[ 3, 0, 2]);
      C( Cell[ 2, 3, 1], Cell[ 3, 2, 0]);
      C( Cell[ 2, 3, 1], Cell[ 3, 2, 2]);

      C( Cell[ 2, 3, 2], Cell[ 5, 0, 1]);
      C( Cell[ 2, 3, 2], Cell[ 5, 1, 0]);
      C( Cell[ 2, 3, 2], Cell[ 5, 1, 1]);

      C( Cell[ 2, 3, 3], Cell[ 3, 2, 0]);
      C( Cell[ 2, 3, 3], Cell[ 3, 2, 2]);
      C( Cell[ 2, 3, 3], Cell[ 5, 1, 0]);
      C( Cell[ 2, 3, 3], Cell[ 5, 1, 1]);

      // ---- Face 3 --------------------

      C( Cell[ 3, 0, 0], Cell[ 0, 3, 1]);
      C( Cell[ 3, 0, 0], Cell[ 0, 3, 3]);
      C( Cell[ 3, 0, 0], Cell[ 2, 1, 1]);
      C( Cell[ 3, 0, 0], Cell[ 2, 1, 3]);

      C( Cell[ 3, 0, 1], Cell[ 0, 1, 3]);
      C( Cell[ 3, 0, 1], Cell[ 0, 3, 1]);
      C( Cell[ 3, 0, 1], Cell[ 0, 3, 3]);

      C( Cell[ 3, 0, 2], Cell[ 2, 1, 1]);
      C( Cell[ 3, 0, 2], Cell[ 2, 1, 3]);
      C( Cell[ 3, 0, 2], Cell[ 2, 3, 1]);

      // --------------------------------

      C( Cell[ 3, 1, 0], Cell[ 0, 1, 1]);
      C( Cell[ 3, 1, 0], Cell[ 0, 1, 3]);
      C( Cell[ 3, 1, 0], Cell[ 0, 3, 1]);

      C( Cell[ 3, 1, 1], Cell[ 0, 1, 1]);
      C( Cell[ 3, 1, 1], Cell[ 0, 1, 3]);
      C( Cell[ 3, 1, 1], Cell[ 4, 0, 0]);
      C( Cell[ 3, 1, 1], Cell[ 4, 0, 2]);

      C( Cell[ 3, 1, 3], Cell[ 4, 0, 0]);
      C( Cell[ 3, 1, 3], Cell[ 4, 0, 2]);
      C( Cell[ 3, 1, 3], Cell[ 4, 2, 0]);

      // --------------------------------

      C( Cell[ 3, 2, 0], Cell[ 2, 1, 3]);
      C( Cell[ 3, 2, 0], Cell[ 2, 3, 1]);
      C( Cell[ 3, 2, 0], Cell[ 2, 3, 3]);

      C( Cell[ 3, 2, 2], Cell[ 2, 3, 1]);
      C( Cell[ 3, 2, 2], Cell[ 2, 3, 3]);
      C( Cell[ 3, 2, 2], Cell[ 5, 1, 1]);
      C( Cell[ 3, 2, 2], Cell[ 5, 1, 3]);

      C( Cell[ 3, 2, 3], Cell[ 5, 1, 1]);
      C( Cell[ 3, 2, 3], Cell[ 5, 1, 3]);
      C( Cell[ 3, 2, 3], Cell[ 5, 3, 1]);

      // --------------------------------

      C( Cell[ 3, 3, 1], Cell[ 4, 0, 2]);
      C( Cell[ 3, 3, 1], Cell[ 4, 2, 0]);
      C( Cell[ 3, 3, 1], Cell[ 4, 2, 2]);

      C( Cell[ 3, 3, 2], Cell[ 5, 1, 3]);
      C( Cell[ 3, 3, 2], Cell[ 5, 3, 1]);
      C( Cell[ 3, 3, 2], Cell[ 5, 3, 3]);

      C( Cell[ 3, 3, 3], Cell[ 4, 2, 0]);
      C( Cell[ 3, 3, 3], Cell[ 4, 2, 2]);
      C( Cell[ 3, 3, 3], Cell[ 5, 3, 1]);
      C( Cell[ 3, 3, 3], Cell[ 5, 3, 3]);

      // ---- Face 4 --------------------

      C( Cell[ 4, 0, 0], Cell[ 0, 1, 0]);
      C( Cell[ 4, 0, 0], Cell[ 0, 1, 1]);
      C( Cell[ 4, 0, 0], Cell[ 3, 1, 1]);
      C( Cell[ 4, 0, 0], Cell[ 3, 1, 3]);

      C( Cell[ 4, 0, 1], Cell[ 0, 0, 1]);
      C( Cell[ 4, 0, 1], Cell[ 0, 1, 0]);
      C( Cell[ 4, 0, 1], Cell[ 0, 1, 1]);

      C( Cell[ 4, 0, 2], Cell[ 3, 1, 1]);
      C( Cell[ 4, 0, 2], Cell[ 3, 1, 3]);
      C( Cell[ 4, 0, 2], Cell[ 3, 3, 1]);

      // --------------------------------

      C( Cell[ 4, 1, 0], Cell[ 0, 0, 0]);
      C( Cell[ 4, 1, 0], Cell[ 0, 0, 1]);
      C( Cell[ 4, 1, 0], Cell[ 0, 1, 0]);

      C( Cell[ 4, 1, 1], Cell[ 0, 0, 0]);
      C( Cell[ 4, 1, 1], Cell[ 0, 0, 1]);
      C( Cell[ 4, 1, 1], Cell[ 1, 0, 0]);
      C( Cell[ 4, 1, 1], Cell[ 1, 0, 2]);

      C( Cell[ 4, 1, 3], Cell[ 1, 0, 0]);
      C( Cell[ 4, 1, 3], Cell[ 1, 0, 2]);
      C( Cell[ 4, 1, 3], Cell[ 1, 2, 0]);

      // --------------------------------

      C( Cell[ 4, 2, 0], Cell[ 3, 1, 3]);
      C( Cell[ 4, 2, 0], Cell[ 3, 3, 1]);
      C( Cell[ 4, 2, 0], Cell[ 3, 3, 3]);

      C( Cell[ 4, 2, 2], Cell[ 3, 3, 1]);
      C( Cell[ 4, 2, 2], Cell[ 3, 3, 3]);
      C( Cell[ 4, 2, 2], Cell[ 5, 3, 2]);
      C( Cell[ 4, 2, 2], Cell[ 5, 3, 3]);

      C( Cell[ 4, 2, 3], Cell[ 5, 2, 3]);
      C( Cell[ 4, 2, 3], Cell[ 5, 3, 2]);
      C( Cell[ 4, 2, 3], Cell[ 5, 3, 3]);

      // --------------------------------

      C( Cell[ 4, 3, 1], Cell[ 1, 0, 2]);
      C( Cell[ 4, 3, 1], Cell[ 1, 2, 0]);
      C( Cell[ 4, 3, 1], Cell[ 1, 2, 2]);

      C( Cell[ 4, 3, 2], Cell[ 5, 2, 2]);
      C( Cell[ 4, 3, 2], Cell[ 5, 2, 3]);
      C( Cell[ 4, 3, 2], Cell[ 5, 3, 2]);

      C( Cell[ 4, 3, 3], Cell[ 1, 2, 0]);
      C( Cell[ 4, 3, 3], Cell[ 1, 2, 2]);
      C( Cell[ 4, 3, 3], Cell[ 5, 2, 2]);
      C( Cell[ 4, 3, 3], Cell[ 5, 2, 3]);

      // ---- Face 5 --------------------

      C( Cell[ 5, 0, 0], Cell[ 1, 3, 2]);
      C( Cell[ 5, 0, 0], Cell[ 1, 3, 3]);
      C( Cell[ 5, 0, 0], Cell[ 2, 2, 2]);
      C( Cell[ 5, 0, 0], Cell[ 2, 2, 3]);

      C( Cell[ 5, 0, 1], Cell[ 2, 2, 2]);
      C( Cell[ 5, 0, 1], Cell[ 2, 2, 3]);
      C( Cell[ 5, 0, 1], Cell[ 2, 3, 2]);

      C( Cell[ 5, 0, 2], Cell[ 1, 2, 3]);
      C( Cell[ 5, 0, 2], Cell[ 1, 3, 2]);
      C( Cell[ 5, 0, 2], Cell[ 1, 3, 3]);

      // --------------------------------

      C( Cell[ 5, 1, 0], Cell[ 2, 2, 3]);
      C( Cell[ 5, 1, 0], Cell[ 2, 3, 2]);
      C( Cell[ 5, 1, 0], Cell[ 2, 3, 3]);

      C( Cell[ 5, 1, 1], Cell[ 2, 3, 2]);
      C( Cell[ 5, 1, 1], Cell[ 2, 3, 3]);
      C( Cell[ 5, 1, 1], Cell[ 3, 2, 2]);
      C( Cell[ 5, 1, 1], Cell[ 3, 2, 3]);

      C( Cell[ 5, 1, 3], Cell[ 3, 2, 2]);
      C( Cell[ 5, 1, 3], Cell[ 3, 2, 3]);
      C( Cell[ 5, 1, 3], Cell[ 3, 3, 2]);

      // --------------------------------

      C( Cell[ 5, 2, 0], Cell[ 1, 2, 2]);
      C( Cell[ 5, 2, 0], Cell[ 1, 2, 3]);
      C( Cell[ 5, 2, 0], Cell[ 1, 3, 2]);

      C( Cell[ 5, 2, 2], Cell[ 1, 2, 2]);
      C( Cell[ 5, 2, 2], Cell[ 1, 2, 3]);
      C( Cell[ 5, 2, 2], Cell[ 4, 3, 2]);
      C( Cell[ 5, 2, 2], Cell[ 4, 3, 3]);

      C( Cell[ 5, 2, 3], Cell[ 4, 2, 3]);
      C( Cell[ 5, 2, 3], Cell[ 4, 3, 2]);
      C( Cell[ 5, 2, 3], Cell[ 4, 3, 3]);

      // --------------------------------

      C( Cell[ 5, 3, 1], Cell[ 3, 2, 3]);
      C( Cell[ 5, 3, 1], Cell[ 3, 3, 2]);
      C( Cell[ 5, 3, 1], Cell[ 3, 3, 3]);

      C( Cell[ 5, 3, 2], Cell[ 4, 2, 2]);
      C( Cell[ 5, 3, 2], Cell[ 4, 2, 3]);
      C( Cell[ 5, 3, 2], Cell[ 4, 3, 2]);

      C( Cell[ 5, 3, 3], Cell[ 3, 3, 2]);
      C( Cell[ 5, 3, 3], Cell[ 3, 3, 3]);
      C( Cell[ 5, 3, 3], Cell[ 4, 2, 2]);
      C( Cell[ 5, 3, 3], Cell[ 4, 2, 3]);

      // --------------------------------
    end;


//  protected

    function    TCubeLife.Getbackground: string;
    begin
      if   Assigned( FOnGetBackground)
      then Result := FOnGetBackground( Self)
      else Result := '';
    end;


//  public

    constructor TCubeLife.Create( aHistorySize: Integer);
    var
      i : Integer;
    begin
      FHistory := TCubeHistory.Create( aHistorySize);

      for i := Low( FFaces) to High( FFaces)
      do FFaces[ i] := TCubeFace.Create( Self, i);

      FixLocalConnections;
    end;


    destructor  TCubeLife.Destroy; // override;
    var
      i : Integer;
    begin
      for i := Low( FFaces) to High( FFaces)
      do FreeAndNil( FFaces[ i]);

      FreeAndNil( FHistory);
      inherited;
    end;


    procedure   TCubeLife.Process;
    var
      i : Integer;
    begin
      FHistory.Push( Pattern);

      for i := Low( FFaces) to High( FFaces)
      do FFaces[ i].Process( UseRedRule);

      for i := Low( FFaces) to High( FFaces)
      do FFaces[ i].Proceed;
    end;


    procedure   TCubeLife.Unprocess;
    begin
      if   FHistory.FillCount > 0
      then Pattern := FHistory.Pop;
    end;


    function    TCubeLife.Dump: string;
    var
      i : Integer;
    begin
      Result := '';

      for i := Low( FFaces) to High( FFaces)
      do Result := Result + Format( '%s'^M, [ FFaces[ i].Dump]);
    end;


    procedure   TCubeLife.PaintFaceOn( aFace: Integer; const aCanvas: TCanvas; const aBounds: TRect);
    begin
      FFaces[ aFace].PaintOn( aCanvas, aBounds);
    end;


{$ifdef USE_FMX}

    procedure   TCubeLife.PaintOn( const aCanvas: TCanvas; const aBounds: TRect);
    var
      Aspect    : Double;
      PanelSize : Integer;
      aBitmap   : TBitmap;
    begin
      if   Assigned( aCanvas)
      then begin
        Aspect := aBounds.Width / aBounds.Height;

        if   Aspect > 4/3
        then aBounds.Width := aBounds.Height * 4 div 3
        else if Aspect < 4/3
        then aBounds.Height := aBounds.Width * 3 div 4;

        PanelSize := aBounds.Width div 4;
        aBitmap   := TBitmap.Create;

        try
          aBitmap.Width  := PanelSize;
          aBitmap.Height := PanelSize;
          aCanvas.BeginScene;
          PaintFaceOn( 0, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 1/4, PanelSize * 0/4, PanelSize * 2/4, PanelSize * 1/3), 1.0);
          PaintFaceOn( 1, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 0/4, PanelSize * 1/3, PanelSize * 1/4, PanelSize * 2/3), 1.0);
          PaintFaceOn( 2, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 1/4, PanelSize * 1/3, PanelSize * 2/4, PanelSize * 2/3), 1.0);
          PaintFaceOn( 3, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 2/4, PanelSize * 1/3, PanelSize * 3/4, PanelSize * 2/3), 1.0);
          PaintFaceOn( 4, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 3/4, PanelSize * 1/3, PanelSize * 4/4, PanelSize * 2/3), 1.0);
          PaintFaceOn( 5, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.DrawBitmap( aBitmap, RectF( PanelSize * 0, PanelSize * 0, PanelSize * 1, PanelSize * 1), RectF( PanelSize * 1/4, PanelSize * 2/3, PanelSize * 2/4, PanelSize * 3/3), 1.0);
          aCanvas.EndScene;
        finally
          aBitmap.DisposeOf;
        end;
      end;
    end;

{$else}

    procedure   TCubeLife.PaintOn( const aCanvas: TCanvas; const aBounds: TRect);
    var
      Aspect    : Double;
      PanelSize : Integer;
      aBitmap   : TBitmap;
    begin
      if   Assigned( aCanvas)
      then begin
        Aspect := aBounds.Width / aBounds.Height;

        if   Aspect > 4/3
        then aBounds.Width := aBounds.Height * 4 div 3
        else if Aspect < 4/3
        then aBounds.Height := aBounds.Width * 3 div 4;

        PanelSize := aBounds.Width div 4;
        aBitmap   := TBitmap.Create;

        try
          aBitmap.Width  := PanelSize;
          aBitmap.Height := PanelSize;

          PaintFaceOn( 0, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.Draw(     PanelSize, 0            , aBitmap);
          PaintFaceOn( 1, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.Draw( 0            ,     PanelSize, aBitmap);
          PaintFaceOn( 2, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.Draw(     PanelSize,     PanelSize, aBitmap);
          PaintFaceOn( 3, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.Draw( 2 * PanelSize,     PanelSize, aBitmap);
          PaintFaceOn( 4, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.Draw( 3 * PanelSize,     PanelSize, aBitmap);
          PaintFaceOn( 5, aBitmap.Canvas, Rect( 0, 0, PanelSize, PanelSize)); aCanvas.Draw(     PanelSize, 2 * PanelSize, aBitmap);
        finally
          aBitmap.DisposeOf;
        end;
      end;
    end;

{$endif}

    procedure   TCubeLife.Clear;
    var
      i : Integer;
    begin
      for i := Low( FFaces) to High( FFaces)
      do FFaces[ i].Clear;
    end;


    function    TCubeLife.MouseToCell( anX, anY: Integer; const aBounds: TRect): TCubeCell;
    var
      PanelSize : Integer;
      R         : TRect;
    begin
      Result := nil;

      if   aBounds.Contains( Point( anX, anY))
      then begin
        PanelSize := aBounds.Width div 4;
        R         := Rect( 0, 0, PanelSize, PanelSize);
        Result    := FFaces[ 0].MouseToCell( anX - 1 * PanelSize, anY - 0 * PanelSize, R); if Assigned( Result) then Exit;
        Result    := FFaces[ 1].MouseToCell( anX - 0 * PanelSize, anY - 1 * PanelSize, R); if Assigned( Result) then Exit;
        Result    := FFaces[ 2].MouseToCell( anX - 1 * panelSize, anY - 1 * PanelSize, R); if Assigned( Result) then Exit;
        Result    := FFaces[ 3].MouseToCell( anX - 2 * panelSize, anY - 1 * PanelSize, R); if Assigned( Result) then Exit;
        Result    := FFaces[ 4].MouseToCell( anX - 3 * panelSize, anY - 1 * PanelSize, R); if Assigned( Result) then Exit;
        Result    := FFaces[ 5].MouseToCell( anX - 1 * panelSize, anY - 2 * PanelSize, R); if Assigned( Result) then Exit;
      end;
    end;


    function    TCubeLife.MouseToFaceCell( aFace: Integer; anX, anY: Integer; const aBounds: TRect): TCubeCell;
    begin
      Result := FFaces[ aFace].MouseToCell( anX, anY, aBounds);
    end;


    procedure   TCubeLife.LoadPattern( const aFileName: string);
    var
      aFile : TextFile;
      S     : string;
    begin
      if   FileExists( aFileName)
      then begin
        AssignFile( aFile, aFileName);
        Reset( aFile);

        try
          ReadLn( aFile, S);
          Pattern := S;
        finally
          CloseFile( aFile);
        end;
      end;
    end;


    procedure   TCubeLife.SavePattern( const aFileName: string);
    var
      aFile : TextFile;
    begin
      AssignFile( aFile, aFileName);
      Rewrite( aFile);

      try
        WriteLn( aFile, Pattern);
      finally
        CloseFile( aFile);
      end;
    end;


    procedure   TCubeLife.ResetHistory;
    begin
      FHistory.Clear;
    end;


end.
